using System;
using System.Runtime.InteropServices;
using System.Threading;
using CoreAnimation;
using CoreGraphics;
using Foundation;
using Metal;
using OpenTK;
using System.Collections.Generic;

namespace VixarEngine
{
	public class Renderer
	{
		const int MaxInflightBuffers = 3;

		nuint _sampleCount;
		IMTLDevice _device;
		MTLPixelFormat _depthPixelFormat;
		MTLPixelFormat _stencilPixelFormat;
	    List<MetalModel3D> _models = new List<MetalModel3D>();

		// Interface Orientation
//		UIInterfaceOrientation mnOrientation;
		Semaphore _inflightSemaphore;

		// Renderer globals
		IMTLCommandQueue _commandQueue;
		IMTLLibrary _shaderLibrary;
		IMTLDepthStencilState _depthState;

		// textured Quad
		IMTLRenderPipelineState _pipelineState;

		Matrix4 _transform;
		IMTLBuffer _transformBuffer;
        Vector3 _pos = new Vector3(0, 0, 0);
        Vector3 _rotate = new Vector3(0, 0, 0);
		Vector3 _eye = new Vector3(100, 200, 1000);
		Vector3 _center = new Vector3(0, 0, 0);
		Vector3 _up = new Vector3(0, 1, 0);
        public bool _touch = false;

		public Renderer ()
		{
			// initialize properties
			_sampleCount = 1;
			_depthPixelFormat = MTLPixelFormat.Depth32Float;
			_stencilPixelFormat = MTLPixelFormat.Invalid;

			// find a usable Device
			_device = MTLDevice.SystemDefault;

			// create a new command queue
			_commandQueue = _device.CreateCommandQueue ();

			NSError error;
			_shaderLibrary = _device.CreateLibrary ("default.metallib", out error);

			// if the shader libary isnt loading, nothing good will happen
			if (_shaderLibrary == null)
				throw new Exception ("ERROR: Couldnt create a default shader library");

			_inflightSemaphore = new Semaphore (MaxInflightBuffers, MaxInflightBuffers);
  		}

		public void Encode (IMTLRenderCommandEncoder renderEncoder,int index)
		{
			renderEncoder.PushDebugGroup (string.Format("encode model {0}",index));
            renderEncoder.SetCullMode(MTLCullMode.Back);
			renderEncoder.SetDepthStencilState (_depthState);
			renderEncoder.SetRenderPipelineState (_pipelineState);
			renderEncoder.SetVertexBuffer (_transformBuffer, 0, 2);

			// Encode quad vertex and texture coordinate buffers
			_models[index].Encode (renderEncoder);

			renderEncoder.PopDebugGroup ();
		}

		public void Render (ImageView view)
		{
			CreateMatrix((float)(view.Bounds.Size.Width / view.Bounds.Size.Height));
			UpdateBuffer();

			_inflightSemaphore.WaitOne ();
			IMTLCommandBuffer commandBuffer = _commandQueue.CommandBuffer ();

			// compute image processing on the (same) drawable texture
			ICAMetalDrawable drawable = view.GetNextDrawable ();

			// create a render command encoder so we can render into something
			MTLRenderPassDescriptor renderPassDescriptor = view.GetRenderPassDescriptor (drawable);

			if (renderPassDescriptor == null) {
				_inflightSemaphore.Release ();
				return;
			}

			// Get a render encoder
			IMTLRenderCommandEncoder renderEncoder = commandBuffer.CreateRenderCommandEncoder (renderPassDescriptor);

			// render textured quad
            for (int i = 0; i < _models.Count; i++ )
				Encode (renderEncoder,i);
			renderEncoder.EndEncoding();

			commandBuffer.AddCompletedHandler ((IMTLCommandBuffer buffer) => {
				_inflightSemaphore.Release ();
				drawable.Dispose ();
			});

			commandBuffer.PresentDrawable (drawable);
			commandBuffer.Commit ();
		}

        public void Configure (ImageView renderView)
		{
			renderView.DepthPixelFormat = _depthPixelFormat;
			renderView.StencilPixelFormat = _stencilPixelFormat;
			renderView.SampleCount = _sampleCount;

			var metalLayer = (CAMetalLayer)renderView.Layer;
			metalLayer.FramebufferOnly = false;

			if (!PreparePipelineState ())
				throw new ApplicationException ("ERROR: Failed creating a depth stencil state descriptor!");

//            LoadModel(new StopJersey(), "StopJersey");

            LoadModel(new JerseyModel(), "Jersey");

			if (!PrepareDepthStencilState ())
				throw new ApplicationException ("ERROR: Failed creating a depth stencil state!");

			if (!PrepareTransformBuffer ())
				throw new ApplicationException ("ERROR: Failed creating a transform buffer!");

			// Default orientation is unknown
//			mnOrientation = UIInterfaceOrientation.Unknown;
		}

		bool PreparePipelineState ()
		{
			// get the fragment function from the library
			IMTLFunction fragmentProgram = _shaderLibrary.CreateFunction ("texturedQuadFragment");

			if (fragmentProgram == null)
				Console.WriteLine ("ERROR: Couldn't load fragment function from default library");

			// get the vertex function from the library
			IMTLFunction vertexProgram = _shaderLibrary.CreateFunction ("texturedQuadVertex");

			if (vertexProgram == null)
				Console.WriteLine ("ERROR: Couldn't load vertex function from default library");

			//  create a pipeline state for the quad
			var quadPipelineStateDescriptor = new MTLRenderPipelineDescriptor {
				DepthAttachmentPixelFormat = MTLPixelFormat.Depth32Float,
				StencilAttachmentPixelFormat = MTLPixelFormat.Invalid,
				SampleCount = _sampleCount,
				VertexFunction = vertexProgram,
				FragmentFunction = fragmentProgram
			};

			quadPipelineStateDescriptor.ColorAttachments [0].PixelFormat = MTLPixelFormat.BGRA8Unorm;

			NSError error;

			_pipelineState = _device.CreateRenderPipelineState (quadPipelineStateDescriptor, out error);
			if (_pipelineState == null) {
				Console.WriteLine ("ERROR: Failed acquiring pipeline state descriptor: %@", error.Description);
				return false;
			}

			return true;
		}

		bool LoadModel(MetalModel3D model,string label)
		{
            if (!model.Load3DData(_device, label))
                throw new ApplicationException("ERROR: Failed creating a " + label);
			else
                _models.Add(model);

			return true;
		}

		bool PrepareTransformBuffer ()
		{
			// allocate regions of memory for the constant buffer
			_transformBuffer = _device.CreateBuffer ((nuint)Marshal.SizeOf<Matrix4> (), MTLResourceOptions.CpuCacheModeDefault);

			if (_transformBuffer == null)
				return false;

			_transformBuffer.Label = "TransformBuffer";

			return true;
		}

		bool PrepareDepthStencilState ()
		{
			var depthStateDesc = new MTLDepthStencilDescriptor {
                DepthCompareFunction = MTLCompareFunction.Less,
				DepthWriteEnabled = true
			};

			_depthState = _device.CreateDepthStencilState (depthStateDesc);

			if (_depthState == null)
				return false;

			return true;
		}

		void CreateMatrix (float aspect)
		{
			Matrix4 perspective = MathUtils.PerspectiveFov(30f, aspect, 0.1f, 10000.0f);
            Matrix4 lookAt = MathUtils.LookAt(_eye, _center, _up);
            if ( _touch )
            {
                float radian = MathUtils.toRadians(_rotate.Y);
                _pos.X += (float)(Math.Sin(radian)*0.05);
                _pos.Z += (float)(Math.Cos(radian)*0.05);
            }
            else
            {
                _rotate.Y++;
                 if (_rotate.Y >= 360) _rotate.Y -= 360;
            }
            Matrix4 rotation = MathUtils.Rotate(_rotate.Y, new Vector3(0, 1, 0));
            _transform = (perspective * lookAt * rotation).SwapColumnsAndRows();
		}

		void UpdateBuffer ()
		{
			// Update the buffer associated with the linear _transformation matrix
			int rawsize = Marshal.SizeOf <Matrix4> ();
			var rawdata = new byte[rawsize];

			GCHandle pinnedTransform = GCHandle.Alloc (_transform, GCHandleType.Pinned);
			IntPtr ptr = pinnedTransform.AddrOfPinnedObject ();
			Marshal.Copy (ptr, rawdata, 0, rawsize);
			pinnedTransform.Free ();

			Marshal.Copy (rawdata, 0, _transformBuffer.Contents, rawsize);
		}
	}

}
